open Typage;;

(**
 * Excute (1 fois) le programme pass en argument et retourne le rsultat
 * de cette excution (se dplacer ou rester sur place).
 * @param pgm Le programme  excuter
 * @param capteurs L'tat des capteurs.
 * @return Une action (NORD, SUD, EST, OUEST ou STOP) ou une valeur (VRAI ou FAUX)
 *)
let rec exec_prog pgm capteurs = 
  match pgm.code with
      Empty -> EvalActionStop (* Cas d'un programme vide *)
        (* Evaluation des oprateurs *)
    | NodeOpSI (fils1,fils2,fils3) -> (
        match exec_prog fils1 capteurs with 
            EvalBoolVrai -> (
              match exec_prog fils2 capteurs with 
                  EvalBoolVrai
                | EvalBoolFaux -> EvalActionStop
                | a -> a (* action *) 
            )
          | EvalBoolFaux -> (
              match exec_prog fils3 capteurs with
                  EvalBoolVrai
                | EvalBoolFaux -> EvalActionStop
                | a -> a (* action *) 
            )
          | a -> a (* action *)
      )
    | NodeOpOU (fils1,fils2) -> (
        match exec_prog fils1 capteurs with 
            EvalBoolVrai -> EvalBoolVrai
          | EvalBoolFaux -> exec_prog fils2 capteurs 
          | a -> a (* EvalActionType action *) 
      )
    | NodeOpET (fils1,fils2) ->  (
        match exec_prog fils1 capteurs with 
            EvalBoolFaux -> EvalBoolFaux
          | EvalBoolVrai -> exec_prog fils2 capteurs 
          | a -> a (* action *) 
      )
    | NodeOpNON (fils1) ->  (
        match exec_prog fils1 capteurs with 
            EvalBoolFaux -> EvalBoolVrai
          | EvalBoolVrai -> EvalBoolFaux
          | a -> a (* action *) 
      )
	(* Evaluation des actions *)
    | NodeActNord  -> EvalActionNord
    | NodeActOuest -> EvalActionOuest
    | NodeActEst   -> EvalActionEst
    | NodeActSud   -> EvalActionSud
	(* Evaluation des variables *)
    | NodeVarN -> if capteurs.(0) then EvalBoolVrai else EvalBoolFaux
    | NodeVarNO -> if capteurs.(1) then EvalBoolVrai else EvalBoolFaux
    | NodeVarNE -> if capteurs.(2) then EvalBoolVrai else EvalBoolFaux
    | NodeVarO -> if capteurs.(3) then EvalBoolVrai else EvalBoolFaux
    | NodeVarE -> if capteurs.(4) then EvalBoolVrai else EvalBoolFaux
    | NodeVarSO -> if capteurs.(5) then EvalBoolVrai else EvalBoolFaux
    | NodeVarSE -> if capteurs.(6) then EvalBoolVrai else EvalBoolFaux
    | NodeVarS -> if capteurs.(7) then EvalBoolVrai else EvalBoolFaux 
        (* Evaluation des constantes *)
    | NodeConst0 -> EvalBoolFaux
    | NodeConst1 -> EvalBoolVrai 
;;

(**
 * Met  jourr l'tat des capteurs (aprs un mouvement).
 * @param posx La position (en ordonnes)
 * @param posy La position (en abscysses)
 * @param grille La grille
 * @param capteurs Le tableau des capteurs  mettre  jour
 * @return Rien (Fonctionnement par effet de bord)
 *)
let check_capteurs posx posy grille capteurs = 
  capteurs.(0) <- grille.(posx).(posy-1);   (* Capteur n *)
  capteurs.(1) <- grille.(posx-1).(posy-1); (* Capteur no *)
  capteurs.(2) <- grille.(posx+1).(posy-1); (* Capteur ne *)
  capteurs.(3) <- grille.(posx-1).(posy);   (* Capteur o *)
  capteurs.(4) <- grille.(posx+1).(posy);   (* Capteur e *)
  capteurs.(5) <- grille.(posx-1).(posy+1); (* Capteur so *)
  capteurs.(6) <- grille.(posx+1).(posy+1); (* Capteur se *)
  capteurs.(7) <- grille.(posx).(posy+1);   (* Capteur s *)
;;

(**
 * Evalue un programme. C'est  dire excute plusieurs fois le programme
 * jusqu' ce que le robot revienne sur une case dj visite ou qu'il
 * s'arrte et compte le nombre de case adjacentes  un mur qu'il a atteintes.
 * @param grille La grille
 * @param posx La position (en ordonnes)
 * @param posy La position (en abscysses)
 * @param pgm Le programme  valuer
 * @return Son valuation (le nombre de case adjacentes  un mur qu'il a atteintes)
 *)
let eval_prog grille posx posy pgm =
  (* On utilise une fonction terminale rcursive pour viter les problmes de mmoire *)
  let rec eval_aux pgm grille posx posy capteurs parcourues score =
    match exec_prog pgm capteurs with
    | EvalActionNord -> 
        (*Printf.fprintf stdout "NORD ";*)
        if parcourues.(posx).(posy-1) then
          (* On s'arrte l car on est rentr dans un mur ou on est revenu sur nos pas *)
          score 
        else ( (* On a avanc sur une nouvelle case *)
          parcourues.(posx).(posy-1) <- true ;
          check_capteurs posx (posy-1) grille capteurs ;
          if (capteurs.(0) or capteurs.(1) or capteurs.(2) or capteurs.(3) or 
              capteurs.(4) or capteurs.(5) or capteurs.(6) or capteurs.(7)) then 
            (* On a avanc sur une case adjacente au mur *)
            eval_aux pgm grille posx (posy - 1) capteurs parcourues (score + 1)
          else
            eval_aux pgm grille posx (posy-1) capteurs parcourues score
         )
    | EvalActionOuest ->
        (*Printf.fprintf stdout "OUEST ";*)
        if parcourues.(posx-1).(posy) then
          (* On s'arrte l car on est rentr dans un mur ou on est revenu sur nos pas *)
          score
        else ( (* On a avanc sur une nouvelle case *)
          check_capteurs (posx-1) posy grille capteurs;
          parcourues.(posx-1).(posy) <- true;
          if capteurs.(0) or capteurs.(1) or capteurs.(2) or capteurs.(3) or 
            capteurs.(4) or capteurs.(5) or capteurs.(6) or capteurs.(7) then 
            (* On a avanc sur une case adjacente au mur *)
            eval_aux pgm grille (posx-1) posy capteurs parcourues (score+1)
          else
            eval_aux pgm grille (posx-1) posy capteurs parcourues score
         )
    | EvalActionEst ->
        (*Printf.fprintf stdout "EST ";*)
        if parcourues.(posx+1).(posy) then
          (* On s'arrte l car on est rentr dans un mur ou on est revenu sur nos pas *)
          score
        else ( (* On a avanc sur une nouvelle case *)
          check_capteurs (posx+1) posy grille capteurs;
          parcourues.(posx+1).(posy) <- true;
          if capteurs.(0) or capteurs.(1) or capteurs.(2) or capteurs.(3) or 
            capteurs.(4) or capteurs.(5) or capteurs.(6) or capteurs.(7) then 
            (* On a avanc sur une case adjacente au mur *)
            eval_aux pgm grille (posx+1) posy capteurs parcourues (score+1)
          else
            eval_aux pgm grille (posx+1) posy capteurs parcourues score
         )
    | EvalActionSud -> 
        (*Printf.fprintf stdout "SUD ";*)
        if parcourues.(posx).(posy+1) then
          (* On s'arrte l car on est rentr dans un mur ou on est revenu sur nos pas *)
          score 
        else ( (* On a avanc sur une nouvelle case *)
          check_capteurs posx (posy+1) grille capteurs;
          parcourues.(posx).(posy+1) <- true;
          if capteurs.(0) or capteurs.(1) or capteurs.(2) or capteurs.(3) or 
            capteurs.(4) or capteurs.(5) or capteurs.(6) or capteurs.(7) then 
            (* On a avanc sur une case adjacente au mur *)
            eval_aux pgm grille posx (posy+1) capteurs parcourues (score+1)
          else
            eval_aux pgm grille posx (posy+1) capteurs parcourues score
         )
    | _ ->
        (*Printf.fprintf stdout "RIEN ";*)
        score
  in
  if grille.(posx).(posy) then
      0 (* On commence dans le mur *)
  else
      let capteurs = Array.make 8 false in
      check_capteurs posx posy grille capteurs;
      (* Recopie grille dans parcourues *)
      let parcourues = Array.map (Array.copy) grille in
      eval_aux pgm grille posx posy capteurs parcourues 0
;;

(**
 * Evalue la performance d'un programme en excutant celui-ci  partir de plusieurs 
 * positions initiales (tires alatoirement).
 * @param grille La grille sur laquelle le robot doit se dplacer
 * @param largeur La largeur de la grille
 * @param hauteur La hauteur de la grille
 * @param nb_essais Le nombre de positions initiales d'o on fait partir le robot
 * @param pgm Le programme codant le comportement du robot
 * @return Le score du programme. C'est  dire la somme des valeurs retournes par 
 *         eval_prog
 *)
let score_prog grille largeur hauteur nb_essais pgm =
  (* Utilisation d'une fonction rcursive terminale pour viter des pbm de mmoire *)
  let rec score_aux grille largeur hauteur nb_essais pgm score =
    if nb_essais = 0 then
      score
    else
      let posx = (Random.int (largeur-2))+1 in
      let posy = (Random.int (hauteur-2))+1 in
      if grille.(posx).(posy) then (* Position initiale dans un mur *)
        score_aux grille largeur hauteur nb_essais pgm score (* On recommence ce coup *)
      else
        score_aux grille largeur hauteur (nb_essais-1) pgm (score + (eval_prog grille posx posy pgm))
  in
  score_aux grille largeur hauteur nb_essais pgm 0
;;
