open Typage;;
open Eval;;
open Util;;

(***************************************************************************
 * Gnration de la population initiale
 ***************************************************************************)

(**
 * Calcul la proba de tirer un lment de programme (par ex un SI)  une hauteur donne 
 * lors de la construction d'un arbre-programme.
 * Ici les tirages sont quiprobables sauf  la profondeur max ou il est impossible de
 * tirer un oprateur (SI, OU ,ET et NON).
 * @param prof La profondeur maximale des programmes
 * @return Un tableau de probabilit (en /oo)  deux dimensions, la 1re indiquant la
 *         hauteur dans l'arbre et la 2me indiquant la proba de chaque type de cellule.
 *)
let make_proba_1 prof = 
  let proba_prof prof indice=
    if indice = (prof-1) then
      Array.init 19 (function 0|1|2|3|4 -> 0 | 17 -> 938 |18 -> 1000 | i -> ((i-4)*73))
    else
      Array.init 19 (function 0 -> 0 | 1 -> 65 | i -> (65 + (i-1)*55))
  in
  Array.init prof (proba_prof prof)
;;

(**
 * Calcul la proba de tirer un lment de programme (par ex un SI)  une hauteur donne 
 * lors de la construction d'un arbre-programme.
 * Ici les proba sont calcules de faon empirique en fonction de la hauteur dans l'arbre.
 * @param prof La profondeur maximale des programmes
 * @return Un tableau de probabilit (en /oo)  deux dimensions, la 1re indiquant la
 *         hauteur dans l'arbre et la 2me indiquant la proba de chaque type de cellule.
 *)
let make_proba_2 prof = 
  let proba_prof prof' indice =
    if indice = 0 then
      Array.init 19 (function 0 -> 0 | 1 -> 600 | 2 -> 800 | _ -> 1000)
    else (
      if indice = prof' then (
        let calcul1 tab  =
          (* Pour NON et ttes las var *)
          for i = 4 to 12 do 
            tab.(i) <- tab.(i-1) + (400/9);
          done;
          (* Pour ttes les actions *)
          for i = 13 to 16 do
            tab.(i) <- tab.(i-1) + (500/4);
          done;
          (* Pour ttes les constantes *)
          tab.(17) <- 950;
          tab.(18) <- 1000;
        in
        let tab = Array.make 19 0 in
        calcul1 tab;
        tab
      )
      else (
        let calcul2 tab t prof' =
          let approx = (if (prof' mod 2) = 1 then prof'/2+1 else prof'/2) in
          (* Pour SI *)
          tab.(1) <-  400-200*t/approx;
          (* Pour ET et OU *)
          for i = 2 to 3 do
            tab.(i) <- tab.(i-1) + (300-100*t/approx)/2;
          done;
          (* Pour NON et ttes las var *)
          for i = 4 to 12 do
            tab.(i) <- tab.(i-1) + (250/9);
          done;
          (* Pour ttes les actions *)
          for i = 13 to 16 do
            tab.(i) <- tab.(i-1) + (100+100*t/approx)/4;
          done;
          (* Pour ttes les constantes *)
          tab.(17) <- tab.(16) + (200*t/approx-50)/2;
          tab.(18) <- 1000;
        in
        let tab = Array.make 19 0 in
        calcul2 tab indice prof';
        tab
      )
    )
  in
  Array.init prof (proba_prof (prof-1))
;;

(**
 * Construit un programme alatoire.
 * Rem : La valeur de la hauteur des sous-arbres dont chaque lment est racine n'est pas
 *       correctement initialise pour les oprateurs.
`* @param proba Tableau contenant pour chaque lment de programme les proba de le tirer
 *         toutes les hauteurs
 * @return Un programme alatoire.
 * @raise EltProgInconnu si un lment choisit alatoirement est de type inconnu.
 *)
let prog_aleat proba =
  let rec prog_aleat_aux proba prof =
    match (quick_find (Random.int 1000) (proba.(prof))) with
        1 -> (* SI *)
          {code = NodeOpSI ((prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))),
                            (prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))),
                            (prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))));
           haut = 1}
      | 2 -> (* ET *)
          {code = NodeOpET ((prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))),
                            (prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))));
           haut = 1}
      | 3 -> (* OU *)
          {code = NodeOpOU ((prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))),
                            (prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))));
           haut = 1}
      | 4 -> (* NON *)
          {code = NodeOpNON ((prog_aleat_aux proba (min (prof+1) ((Array.length proba)-1))));
           haut = 1}
      | 5 ->  (* n  *) {code = NodeVarN ; haut = 1}
      | 6 ->  (* no *) {code = NodeVarNO ; haut = 1}
      | 7 ->  (* ne *) {code = NodeVarNE ; haut = 1}
      | 8 ->  (* o  *) {code = NodeVarO ; haut = 1}
      | 9 ->  (* e  *) {code = NodeVarE ; haut = 1}
      | 10 -> (* so *) {code = NodeVarSO ; haut = 1}
      | 11 -> (* se *) {code = NodeVarSE ; haut = 1}
      | 12 -> (* s *)  {code = NodeVarS ; haut = 1}
      | 13 -> (* Nord  *) {code = NodeActNord ; haut = 1}
      | 14 -> (* Ouest *) {code = NodeActOuest ; haut = 1}
      | 15 -> (* Est   *) {code = NodeActEst ; haut = 1}
      | 16 -> (* Sud   *) {code = NodeActSud ; haut = 1}
      | 17 -> (* 0 *) {code = NodeConst0 ; haut = 1}
      | 18 -> (* 1 *) {code = NodeConst1 ; haut = 1}
      | i -> (* Erreur *) raise (EltProgInconnu i)
  in
  prog_aleat_aux proba 0
;;

(**
 * Gnre alatoirement la premire population de programmes.
 * @param nb_pgm Le nombre de programmes constituant cette gnration
 * @param prof_init La profondeur maximale des premiers programmes
 * @param type_pgm_init Le type de proba pour construire alatoirement les programmes
 * @return Un tableau contenant tous les programme de la premire population
 *)
let init_gen nb_pgm prof_init type_pgm_init = 
  let proba1 = make_proba_1 prof_init in
  let proba2 = make_proba_2 prof_init in
  if type_pgm_init = 1 then
    Array.init nb_pgm (fun i -> (prog_aleat proba1))
  else
    Array.init nb_pgm (fun i -> (prog_aleat proba2))
;;


(***************************************************************************
 * Passage d'une gnration  l'autre
 ***************************************************************************)

(**
 * Droulement d'un tournoi:
 *   - tirage alatoire de nb_parts programmes diffrents
 *   - slection de celui dont le score est le plus lv
 * @param nb_parts Nombre de participants lors d'un tournoi
 * @param scores Tableau contenant le score de tous les programmes de la gnration
 * @return L'indice (dans le tableau des programme) du programme vainqueur
 *)
let tournoi nb_parts scores = 
  (* Ajoute nouveau dans la liste trie en respectant ce tri.
     Si nouveau existe dj dans la liste -> exception ExisteDejaDansListe *)
  let rec ajouter nouveau liste =
    match liste with
        [] -> nouveau::[]
      | head::tail ->
          if head = nouveau then (* le nouveau existe dj dans la liste *)
            raise ExisteDejaDansListe
          else
            if head > nouveau then
              nouveau::liste
            else
              head::(ajouter nouveau tail)
  in
  (* Constitue la liste des participants au tournoi *)
  let rec select_participants nb_pgms nb_restants deja_select =
    if nb_restants = 0 then
      deja_select
    else
      try
        select_participants nb_pgms (nb_restants-1) (ajouter (Random.int nb_pgms) deja_select)
      with
          ExisteDejaDansListe -> select_participants nb_pgms nb_restants deja_select
  in
  (* Slectionne le meilleur prog parmi les participants *)
  let rec meilleur liste max indice_max =
    match liste with
        [] -> indice_max
      | (i,score)::tail ->
          if score > max then 
            meilleur tail score i
          else
            meilleur tail max indice_max
  in
  let nb_pgms = Array.length scores in
  let liste = select_participants nb_pgms nb_parts [] in
  let liste' = List.map (fun i -> (i,scores.(i))) liste in
  (*List.iter (fun (i,s) -> Printf.fprintf stdout "(%d,%d)\n" i s) liste';*)
  meilleur liste' (-1) 0
;;

(**
 * Effectue un tournoi et recopie le vainqueur.
 * @param nb_parts Nombre de participants aux tournois
 * @param scores Tableau des scores des programmes
 * @param pop_cur Tableau contenant les programme de la population actuelle
 * @return La copie d'un programme vainqueur d'un tournoi
 *)
let selection nb_parts scores pop_cur =
  let elu = tournoi nb_parts scores in
  copie_prog (pop_cur.(elu))
;;

(**
 * Extrait une partie de l'arbre-programme pass en paramtre.
 * Le choix de la partie  recopier se fait en 3 tapes:
 * - tirage (selon des probabilits empiriques) d'une hauteur de l'arbre
 * - recherche de tous les noeuds se trouvant  cette hauteur
 * - tiraged'un noeud parmi ceux-ci (plus l'arbre dont ces noeuds sont racine est haut
 *   plus le noeud a des chances d'tre tir)
 * Attention: La partie retourne n'est PAS une copie.
 * @param pgm Le programme
 * @return Une partie du programme pass en paramtre
 *)
let sous_prog pgm =
  let haut = pgm.haut in
  (* Constitution du tableau des proba pour tirer la hauteur du ss-prog *)
  let proba = Array.make (haut + 1) 0 in
  (match haut with
      0 -> raise PgmVide
    | 1 -> proba.(1) <- 1000;
    | 2 ->
        proba.(1) <- 500;
        proba.(2) <- 1000;
    | 3 ->
        proba.(1) <- 500;
        proba.(2) <- 750;
        proba.(3) <- 1000;
    | _ ->
        let p1 = 270 in
        let p2 = 330 in
        let p3 = 400 in
        for h = 1 to (haut/4) do
          proba.(h) <- h * (p1 / (haut/4));
        done;
        for h = (haut/4)+1 to (haut/2) do
          proba.(h) <- p1 + (h-(haut/4)) * (p2 / ((haut/2)-(haut/4)));
        done;
        for h = (haut/2)+1 to (haut-1) do
          proba.(h) <- p1 + p2 + (h-(haut/2)) * (p3 / (haut - (haut/2)));
        done;
        proba.(haut) <- 1000;
  );
  (* Tirage d'une hauteur *)
  let h = (quick_find (Random.int 1000) proba) in
  (* Extraction de tous les noeuds  cette hauteur *)
  let rec extract pgm h_visee h_atteinte = 
    if h_visee = h_atteinte  then
      [pgm]
    else 
      match pgm.code with
          Empty
        | NodeVarN
        | NodeVarNO
        | NodeVarNE
        | NodeVarO
        | NodeVarE
        | NodeVarSO
        | NodeVarSE
        | NodeVarS
        | NodeActNord
        | NodeActOuest
        | NodeActEst
        | NodeActSud
        | NodeConst0
        | NodeConst1 -> []
        | NodeOpSI (fils1,fils2,fils3) -> 
            (if (fils1.haut + h_atteinte) < h_visee then [] else extract fils1 h_visee (h_atteinte+1)) @
            (if (fils2.haut + h_atteinte) < h_visee then [] else extract fils2 h_visee (h_atteinte+1)) @
            (if (fils3.haut + h_atteinte) < h_visee then [] else extract fils3 h_visee (h_atteinte+1))
        | NodeOpOU (fils1,fils2) -> 
            (if (fils1.haut + h_atteinte) < h_visee then [] else extract fils1 h_visee (h_atteinte+1)) @
            (if (fils2.haut + h_atteinte) < h_visee then [] else extract fils2 h_visee (h_atteinte+1))
        | NodeOpET (fils1,fils2) -> 
            (if (fils1.haut + h_atteinte) < h_visee then [] else extract fils1 h_visee (h_atteinte+1)) @
            (if (fils2.haut + h_atteinte) < h_visee then [] else extract fils2 h_visee (h_atteinte+1))
        | NodeOpNON (fils1) -> 
            (if (fils1.haut + h_atteinte) < h_visee then [] else extract fils1 h_visee (h_atteinte+1))
  in
  (* Slection d'un noeud  cette hauteur *)
  let select list_pgm =
    (* Calcul de la probabilit de tirer un prog de cette liste *)
    let rec somme = function
        [] -> 0
      | h::t -> h.haut + (somme t)
    in
    let s = somme list_pgm in
    let proba = Array.make (1 + List.length list_pgm) 0 in
    let rec calcul_proba tab i liste =
      match liste with
          [] -> ignore 1;
        | h::t -> 
            tab.(i) <- (tab.(i-1) + h.haut);
            calcul_proba tab (i+1) t;
    in
    calcul_proba proba 1 list_pgm;
    (* Tirage d'un prog *)
    List.nth list_pgm ((quick_find (Random.int s) proba) - 1) 
  in
  select (extract pgm h 1)
;;

(**
 * Effectue le croisement entre deux programmes tirs au hasard. De ces deux programmes, 
 * il ne rsulte qu'un seul fils.
 * Si p1 et p2 sont les sous programmes  croiser, alors on va greffer  la place d'une 
 * branche de p1 un sous-arbre sp2 de p2. Ce sous-arbre est obtenu grce  la fonction 
 * sous_prog. Pour choisir l'endroit o greffer sp2 dans p1, on utilise un algorithme 
 * glouton probabiliste.
 * @param nb_parts Nombre de participants pendant les tournois pour slectionner les 
 *                 programmes  croiser.
 * @param pop_cur Le tableau des programmes de la gnration actuelle
 * @param scores Le tableau des scores de la gnration actuelle (pour les tournois)
 * @return Un nouveau programme (avec son espace mmoire propre) croisement de p1 et p2
 *)
let croisement nb_parts pop_cur scores =
  let rec croisement_aux pgm greffe =
    let p = 250 in
    match pgm.code with
        Empty
      | NodeVarN
      | NodeVarNO
      | NodeVarNE
      | NodeVarO
      | NodeVarE
      | NodeVarSO
      | NodeVarSE
      | NodeVarS
      | NodeActNord
      | NodeActOuest
      | NodeActEst
      | NodeActSud
      | NodeConst0
      | NodeConst1 -> greffe
      | NodeOpSI (fils1,fils2,fils3) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            (* On place la greffe dans un fils *)
            match (Random.int 3) with (* Tirage du fils dans lequel on va placer la greffe *)
                0 -> (* dans le fils1 *)
                  {code=NodeOpSI(croisement_aux fils1 greffe,copie_prog fils2,copie_prog fils3);haut=1}
              | 1 -> (* dans le fils2 *)
                  {code=NodeOpSI(copie_prog fils1,croisement_aux fils2 greffe,copie_prog fils3);haut=1}
              | _ -> (* dans le fils3 *)
                  {code=NodeOpSI(copie_prog fils1,copie_prog fils2,croisement_aux fils3 greffe);haut=1}
          )
      | NodeOpOU (fils1,fils2) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            (* On place la greffe dans un fils *)
            match (Random.int 3) with (* Tirage du fils dans lequel on va placer la greffe *)
                0 -> (* dans le fils1 *)
                  {code=NodeOpOU(croisement_aux fils1 greffe,copie_prog fils2);haut=1}
              | _ -> (* dans le fils2 *)
                  {code=NodeOpOU(copie_prog fils1,croisement_aux fils2 greffe);haut=1}
          )
      | NodeOpET (fils1,fils2) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            (* On place la greffe dans un fils *)
            match (Random.int 3) with (* Tirage du fils dans lequel on va placer la greffe *)
                0 -> (* dans le fils1 *)
                  {code=NodeOpET(croisement_aux fils1 greffe,copie_prog fils2);haut=1}
              | _ -> (* dans le fils2 *)
                  {code=NodeOpET(copie_prog fils1,croisement_aux fils2 greffe);haut=1}
          )
      | NodeOpNON (fils1) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else
            (* On place la greffe dans le fils *)
            {code=NodeOpNON (croisement_aux fils1 greffe) ; haut=1}
  in
  croisement_aux pop_cur.(tournoi nb_parts scores) (copie_prog (sous_prog (pop_cur.(tournoi nb_parts scores))))
;;

(**
 * Calcule la population suivante  partir de la population actuelle. Cela se droule en 
 * deux tapes:
 *   - calcul du score des programmes de la gnration actuelle
 *   - cration de la gnration suivante qui est compose de:
 *      ~ programmes immortels (recopis sans modification de la gnration actuelle)
 *      ~ programmes enfants (fruit du croisement de deux programmes actuels)
 * @param pop_cur Tableau contenant les programme de la population actuelle
 * @param nb_immort Nombre de programmes  recopier sans modificatioin
 * @param grille L'univers dans lequel le robot se dplace (pour l'valuation)
 * @param nb_essais Nombre de positions init (pour valuation)
 * @param nb_parts Nombre de participants aux tournois
 * @return Un nouveau tableau contenant la nouvelle population
 *)
let pop_suivante pop_cur nb_immort grille nb_essais  nb_parts =
  (* Calcul des scores *)
  let largeur = Array.length grille in
  let hauteur = Array.length grille.(0) in
  let scores = Array.map (score_prog grille largeur hauteur nb_essais) pop_cur in
  (* Calcul de la nouvelle population *)
  let nb_pgms = Array.length pop_cur in
  let new_pgm nb_pgm nb_immort nb_parts scores indice pgm =
    let periode = nb_pgms / nb_immort in
    if (indice mod periode) = 0 then 
      (* Slectionne un programme immortel *)
      selection nb_parts scores pop_cur
    else
      (* Croisement de deux programmes *)
      croisement nb_parts pop_cur scores
  in
  Array.mapi (new_pgm nb_pgms nb_immort nb_parts scores) pop_cur
;;

(**
 * Le pendant de la fonction croisement  la quelle a t ajoute la mutation.
 * @param nb_parts Nombre de participants pendant les tournois pour slectionner les 
 *                 programmes  croiser.
 * @param pop_cur Le tableau des programmes de la gnration actuelle
 * @param scores Le tableau des scores de la gnration actuelle (pour les tournois)
 * @param proba La probabilit de mutation
 * @return Un nouveau programme (avec son espace mmoire propre) croisement de p1 et p2
 *)
let mutation nb_parts pop_cur scores proba =
  let rec mutation_aux pgm greffe =
    let p = 250 in
    match pgm.code with
        Empty
      | NodeVarN
      | NodeVarNO
      | NodeVarNE
      | NodeVarO
      | NodeVarE
      | NodeVarSO
      | NodeVarSE
      | NodeVarS
      | NodeActNord
      | NodeActOuest
      | NodeActEst
      | NodeActSud
      | NodeConst0
      | NodeConst1 -> greffe
      | NodeOpSI (fils1,fils2,fils3) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            (* On place la greffe dans un fils *)
            match (Random.int 3) with (* Tirage du fils dans lequel on va placer la greffe *)
                0 -> (* dans le fils1 *)
                  {code=NodeOpSI(mutation_aux fils1 greffe,cp_mut_prog fils2 proba,cp_mut_prog fils3 proba);
                   haut=1}
              | 1 -> (* dans le fils2 *)
                  {code=NodeOpSI(cp_mut_prog fils1 proba,mutation_aux fils2 greffe,cp_mut_prog fils3 proba);
                   haut=1}
              | _ -> (* dans le fils3 *)
                  {code=NodeOpSI(cp_mut_prog fils1 proba,cp_mut_prog fils2 proba,mutation_aux fils3 greffe);
                   haut=1}
          )
      | NodeOpOU (fils1,fils2) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            if (Random.int 1000) < proba then (* On modifie ce noeud *)
              (* On place la greffe dans un fils *)
              match (Random.int 3) with (* Tirage du fils ds lequel on va placer la greffe *)
                  0 -> (* dans le fils1 *)
                    {code=NodeOpET(mutation_aux fils1 greffe,cp_mut_prog fils2 proba);haut=1}
                | _ -> (* dans le fils2 *)
                    {code=NodeOpET(cp_mut_prog fils1 proba,mutation_aux fils2 greffe);haut=1}
            else
              (* On place la greffe dans un fils *)
              match (Random.int 3) with (* Tirage du fils ds lequel on va placer la greffe *)
                  0 -> (* dans le fils1 *)
                    {code=NodeOpOU(mutation_aux fils1 greffe,cp_mut_prog fils2 proba);haut=1}
                | _ -> (* dans le fils2 *)
                    {code=NodeOpOU(cp_mut_prog fils1 proba,mutation_aux fils2 greffe);haut=1}
          )
      | NodeOpET (fils1,fils2) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else (
            if (Random.int 1000) < proba then (* On modifie ce noeud *)
              (* On place la greffe dans un fils *)
              match (Random.int 3) with (* Tirage du fils ds lequel on va placer la greffe *)
                  0 -> (* dans le fils1 *)
                    {code=NodeOpOU(mutation_aux fils1 greffe,cp_mut_prog fils2 proba);haut=1}
                | _ -> (* dans le fils2 *)
                    {code=NodeOpOU(cp_mut_prog fils1 proba,mutation_aux fils2 greffe);haut=1}
            else
              (* On place la greffe dans un fils *)
              match (Random.int 3) with (* Tirage du fils ds lequel on va placer la greffe *)
                  0 -> (* dans le fils1 *)
                    {code=NodeOpET(mutation_aux fils1 greffe,cp_mut_prog fils2 proba);haut=1}
                | _ -> (* dans le fils2 *)
                    {code=NodeOpET(cp_mut_prog fils1 proba,mutation_aux fils2 greffe);haut=1}
          )
      | NodeOpNON (fils1) -> 
          if (Random.int 1000) < p then
            (* On place la greffe  cet endroit *)
            greffe
          else
            if (Random.int 1000) < proba then (* On modifie ce noeud *)
              mutation_aux fils1 greffe
            else
              (* On place la greffe dans le fils *)
              {code=NodeOpNON (mutation_aux fils1 greffe) ; haut=1}
  in
  mutation_aux (pop_cur.(tournoi nb_parts scores))
    (cp_mut_prog (sous_prog (pop_cur.(tournoi nb_parts scores))) proba)
;;

(**
 * Le pendant de la fonction pop_suivante  la quelle a t ajoute la mutation.
 * @param pop_cur Tableau contenant les programme de la population actuelle
 * @param nb_immort Nombre de programmes  recopier sans modificatioin
 * @param grille L'univers dans lequel le robot se dplace (pour l'valuation)
 * @param nb_essais Nombre de positions init (pour valuation)
 * @param nb_parts Nombre de participants aux tournois
 * @param proba La probabilit de mutation
 * @return Un nouveau tableau contenant la nouvelle population
 *)
let pop_mut_suivante pop_cur nb_immort grille nb_essais nb_parts proba =
  (* Calcul des scores *)
  let largeur = Array.length grille in
  let hauteur = Array.length grille.(0) in
  let scores = Array.map (score_prog grille largeur hauteur nb_essais) pop_cur in
  (* Calcul de la nouvelle population *)
  let nb_pgms = Array.length pop_cur in
  let new_pgm nb_pgm nb_immort nb_parts scores proba indice pgm =
    let periode = nb_pgms / nb_immort in
    if (indice mod periode) = 0 then 
      (* Slectionne un programme immortel *)
      selection nb_parts scores pop_cur
    else
      (* Croisement de deux programmes *)
      mutation nb_parts pop_cur scores proba
  in
  Array.mapi (new_pgm nb_pgms nb_immort nb_parts scores proba) pop_cur
;;
