open Typage;;

(**
 * On part d'un tableau tri strictement croissant commenant par 0.
 * On cherche l'indice de la case de ce tableau correspondant  une valeur donne
 * c'est  dire l'indice i tel que tab.(i-1)<valeur<=tab.(i).
 * La recherche se fait en temps logarithmique.
 * @param valeur La valeur recherche
 * @param tab Le tableau tri
 * @return l'indice correspondant
 * @raise ValeurHorsTab si la valeur est >=  la plus grande valeur du tableau
 *)
let quick_find valeur tab =
  let rec find_aux valeur tab min max cur =
    if tab.(cur) = valeur then
      cur + 1
    else
      if valeur < tab.(cur) then
        if min = (cur-1) then
          cur
        else
          find_aux valeur tab min cur ((max+min)/2)
      else
        if cur = (max-1) then
          max
        else
          if ((max+min) mod 2) = 0 then
            find_aux valeur tab cur max ((max+min)/2)
          else
            find_aux valeur tab cur max ((1+max+min)/2)
  in
  let max = (Array.length tab) - 1 in
  if tab.(max) > valeur then
    find_aux valeur tab 0 max (max/2)
  else
    raise (ValeurHorsTab (valeur,tab.(max)))
;;

(**
 * Affiche de faon lisible un programme sur la sortie standard.
 * @param pgm Le programme  afficher
 * @return Rien
 *)
let affiche pgm = 
  let rec print_esp n =
    if n > 0 then (
      Printf.fprintf stdout " ";
      print_esp (n-1))
  in
  let rec aff_aux node nb_esp = 
    Printf.fprintf stdout "(";
    ( match node.code with
      Empty -> Printf.fprintf stdout "()\n"
    | NodeVarN -> Printf.fprintf stdout "n"
    | NodeVarNO -> Printf.fprintf stdout "no"
    | NodeVarNE -> Printf.fprintf stdout "ne"
    | NodeVarO -> Printf.fprintf stdout "o"
    | NodeVarE -> Printf.fprintf stdout "e"
    | NodeVarSO -> Printf.fprintf stdout "so"
    | NodeVarSE -> Printf.fprintf stdout "se"
    | NodeVarS -> Printf.fprintf stdout "s"
    | NodeActNord -> Printf.fprintf stdout "nord"
    | NodeActOuest -> Printf.fprintf stdout "ouest"
    | NodeActEst -> Printf.fprintf stdout "est"
    | NodeActSud -> Printf.fprintf stdout "sud"
    | NodeConst0 -> Printf.fprintf stdout "0"
    | NodeConst1 -> Printf.fprintf stdout "1"
    | NodeOpSI (fils1,fils2,fils3) -> 
        Printf.fprintf stdout "SI ";
        aff_aux fils1 (nb_esp + 4);
        Printf.fprintf stdout "\n";
        print_esp (nb_esp + 4);
        aff_aux fils2 (nb_esp + 4);
        Printf.fprintf stdout "\n";
        print_esp (nb_esp + 4);
        aff_aux fils3 (nb_esp + 4);
    | NodeOpOU (fils1,fils2) -> 
        Printf.fprintf stdout "OU ";
        aff_aux fils1 (nb_esp + 4);
        Printf.fprintf stdout "\n";
        print_esp (nb_esp + 4);
        aff_aux fils2 (nb_esp + 4);
    | NodeOpET (fils1,fils2) -> 
        Printf.fprintf stdout "ET ";
        aff_aux fils1 (nb_esp + 4);
        Printf.fprintf stdout "\n";
        print_esp (nb_esp + 4);
        aff_aux fils2 (nb_esp + 4);
    | NodeOpNON (fils1) -> 
        Printf.fprintf stdout "NON ";
        aff_aux fils1 (nb_esp + 5);
     );
    Printf.fprintf stdout ")"
  in
  aff_aux pgm 0;
  Printf.fprintf stdout "\n";
  flush stdout
;;

(**
 * Recopie  entirement un programme (dans un espace mmoire nouvellement allou).
 * @param pgm Le programme  recopier
 * @return Une copie du programme
 *)
let rec copie_prog pgm =
  match pgm.code with
      Empty -> {code = Empty ; haut = 0}
    | NodeVarN -> {code = NodeVarN ; haut = 1}
    | NodeVarNO -> {code = NodeVarNO ; haut = 1}
    | NodeVarNE -> {code = NodeVarNE ; haut = 1}
    | NodeVarO -> {code = NodeVarO ; haut = 1}
    | NodeVarE -> {code = NodeVarE ; haut = 1}
    | NodeVarSO -> {code = NodeVarSO ; haut = 1}
    | NodeVarSE -> {code = NodeVarSE ; haut = 1}
    | NodeVarS -> {code = NodeVarS ; haut = 1}
    | NodeActNord -> {code = NodeActNord ; haut = 1}
    | NodeActOuest -> {code = NodeActOuest ; haut = 1}
    | NodeActEst -> {code = NodeActEst ; haut = 1}
    | NodeActSud -> {code = NodeActSud ; haut = 1}
    | NodeConst0 -> {code = NodeConst0 ; haut = 1}
    | NodeConst1 -> {code = NodeConst1 ; haut = 1}
    | NodeOpSI (fils1,fils2,fils3) -> 
        {code = NodeOpSI (copie_prog fils1,copie_prog fils2,copie_prog fils3) ;
         haut = pgm.haut}
    | NodeOpOU (fils1,fils2) -> 
        {code = NodeOpOU (copie_prog fils1,copie_prog fils2) ;
         haut = pgm.haut}
    | NodeOpET (fils1,fils2) ->
        {code = NodeOpET (copie_prog fils1,copie_prog fils2) ;
         haut = pgm.haut}
    | NodeOpNON (fils1)  ->
        {code = NodeOpNON (copie_prog fils1) ;
         haut = pgm.haut}
;;

(**
 * Calcule la hauteur du programme pass en paramtre et de tous les sous programmes
 * qui le compose.
 * La hauteur du programme est retourne directement par la fonction. La hauteur des
 * sous programmes est stocke dans le noeud racine du sous-programme.
 * @param pgm Le programme dont on veut calculer la hauteur
 * @return La hauteur du programme
 *)
let rec haut_prog pgm =
  match pgm.code with
      Empty ->
        (pgm.haut <- 0);
        0
    | NodeVarN
    | NodeVarNO
    | NodeVarNE
    | NodeVarO
    | NodeVarE
    | NodeVarSO
    | NodeVarSE
    | NodeVarS
    | NodeActNord
    | NodeActOuest
    | NodeActEst
    | NodeActSud
    | NodeConst0
    | NodeConst1 -> 
        (pgm.haut <- 1); 
        1
    | NodeOpSI (fils1,fils2,fils3) -> 
        (pgm.haut <- (1 + (max (max (haut_prog fils1) (haut_prog fils2)) (haut_prog fils3))));
        pgm.haut
    | NodeOpOU (fils1,fils2) -> 
        (pgm.haut <- (1 + (max (haut_prog fils1) (haut_prog fils2))));
        pgm.haut
    | NodeOpET (fils1,fils2) -> 
        (pgm.haut <- (1 + (max (haut_prog fils1) (haut_prog fils2))));
        pgm.haut
    | NodeOpNON (fils1) -> 
        (pgm.haut <- (1 + (haut_prog fils1)));
        pgm.haut
;;

(**
 * Rduit la taille d'un programme en le simplifiant. C'est  dire en supprimant
 * les blocs d'instructions inutils (par exemple ET(0,x) devient 0).
 * @param pgm Le programme  simplifier
 * @return Un nouveau programme (ayant son propre espace mmoire) rduit
 *         et quivalent  pgm
 *)
let rec simpl_prog pgm =
  match pgm.code with
      (* Les feuilles sont recopie sans modification *)
      Empty -> {code = Empty ; haut = 0}
    | NodeVarN -> {code = NodeVarN ; haut = 1}
    | NodeVarNO -> {code = NodeVarNO ; haut = 1}
    | NodeVarNE -> {code = NodeVarNE ; haut = 1}
    | NodeVarO -> {code = NodeVarO ; haut = 1}
    | NodeVarE -> {code = NodeVarE ; haut = 1}
    | NodeVarSO -> {code = NodeVarSO ; haut = 1}
    | NodeVarSE -> {code = NodeVarSE ; haut = 1}
    | NodeVarS -> {code = NodeVarS ; haut = 1}
    | NodeActNord -> {code = NodeActNord ; haut = 1}
    | NodeActOuest -> {code = NodeActOuest ; haut = 1}
    | NodeActEst -> {code = NodeActEst ; haut = 1}
    | NodeActSud -> {code = NodeActSud ; haut = 1}
    | NodeConst0 -> {code = NodeConst0 ; haut = 1}
    | NodeConst1 -> {code = NodeConst1 ; haut = 1}
      (* Les noeuds internes sont ventuellement rduits *)
    | NodeOpSI (fils1,fils2,fils3) -> (
        let f1_simpl = simpl_prog fils1 in
        match f1_simpl.code with
            NodeActNord
          | NodeActOuest
          | NodeActEst
          | NodeActSud
          | NodeOpSI (_,_,_) -> f1_simpl
          | NodeConst1 -> (
              let f2_simpl = simpl_prog fils2 in 
              match f2_simpl.code with
                  NodeActNord
                | NodeActOuest
                | NodeActEst
                | NodeActSud
                | NodeOpSI (_,_,_) -> f2_simpl
                | _ -> {code = NodeOpSI (f1_simpl,f2_simpl,simpl_prog fils3); haut=1}
            )
          | NodeConst0 -> (
              let f3_simpl = simpl_prog fils3 in 
              match f3_simpl.code with
                  NodeActNord
                | NodeActOuest
                | NodeActEst
                | NodeActSud
                | NodeOpSI (_,_,_) -> f3_simpl
                | _ -> {code = NodeOpSI (f1_simpl,simpl_prog fils2,f3_simpl); haut=1}
            )
          | _ -> {code = NodeOpSI (f1_simpl,simpl_prog fils2,simpl_prog fils3); haut=1}
      )
    | NodeOpOU (fils1,fils2) -> (
        let f1_simpl = simpl_prog fils1 in
        match f1_simpl.code with
            NodeActNord
          | NodeActOuest
          | NodeActEst
          | NodeActSud
          | NodeOpSI (_,_,_)
          | NodeConst1 -> f1_simpl
          | NodeConst0 -> simpl_prog fils2
          | NodeVarN
          | NodeVarNO
          | NodeVarNE
          | NodeVarO
          | NodeVarE
          | NodeVarSO
          | NodeVarSE
          | NodeVarS -> (
              let f2_simpl = simpl_prog fils2 in
              match f2_simpl.code with 
                  NodeConst1 -> f2_simpl
                | NodeConst0 -> f1_simpl
                | _ -> {code = NodeOpOU (f1_simpl,f2_simpl) ; haut = 1}
              )
          | _ -> {code = NodeOpOU (f1_simpl,simpl_prog fils2) ; haut = 1}
      )
    | NodeOpET (fils1,fils2) -> (
        let f1_simpl = simpl_prog fils1 in
        match f1_simpl.code with
            NodeActNord
          | NodeActOuest
          | NodeActEst
          | NodeActSud
          | NodeOpSI (_,_,_)
          | NodeConst0 -> f1_simpl
          | NodeConst1 -> simpl_prog fils2
          | NodeVarN
          | NodeVarNO
          | NodeVarNE
          | NodeVarO
          | NodeVarE
          | NodeVarSO
          | NodeVarSE
          | NodeVarS -> (
              let f2_simpl = simpl_prog fils2 in
              match f2_simpl.code with
                  NodeConst0 -> f2_simpl
                | NodeConst1 -> f1_simpl
                | _ -> {code = NodeOpET (f1_simpl,f2_simpl) ; haut = 1}
             )
          | _ -> {code = NodeOpET (f1_simpl,simpl_prog fils2) ; haut = 1}
      )
    | NodeOpNON (fils1) -> (
        let f1_simpl = simpl_prog fils1 in
        match f1_simpl.code with
            NodeActNord
          | NodeActOuest
          | NodeActEst
          | NodeActSud -> f1_simpl
          | NodeConst0 -> {code = NodeConst1 ; haut = 1}
          | NodeConst1 -> {code = NodeConst0 ; haut = 1}
          | NodeOpNON (f) -> f
          | _ -> {code = NodeOpNON (f1_simpl) ; haut = 1}
      )
;;

(**
 * Recopie entirement un programme en modifiant ventuellement quelques noeuds.
 * @param pgm Le programme  recopier/altrer
 * @param proba La probabilit qu'un noeud soit altr
 * @return Une copie du programme (dans un espace mmoire nouvellement allou)
 *)
let rec cp_mut_prog pgm proba =
  match pgm.code with
      Empty -> {code = Empty ; haut = 0}
    | NodeVarN
    | NodeVarNO
    | NodeVarNE
    | NodeVarO
    | NodeVarE
    | NodeVarSO
    | NodeVarSE
    | NodeVarS
    | NodeActNord
    | NodeActOuest
    | NodeActEst
    | NodeActSud
    | NodeConst0
    | NodeConst1 -> (
        if Random.int 1000 < proba then (* On altre le noeud *)
          let p_non = 150 in
          let new_var_act_cst () =
            match Random.int 14 with
                0 -> {code = NodeVarN ; haut = 1}
              | 1 -> {code = NodeVarNO ; haut = 1}
              | 2 -> {code = NodeVarNE ; haut = 1}
              | 3 -> {code = NodeVarO ; haut = 1}
              | 4 -> {code = NodeVarE ; haut = 1}
              | 5 -> {code = NodeVarSO ; haut = 1}
              | 6 -> {code = NodeVarSE ; haut = 1}
              | 7 -> {code = NodeVarS ; haut = 1}
              | 8 -> {code = NodeActNord ; haut = 1}
              | 9 -> {code = NodeActOuest ; haut = 1}
              | 10 -> {code = NodeActEst ; haut = 1}
              | 11 -> {code = NodeActSud ; haut = 1}
              | 12 -> {code = NodeConst0 ; haut = 1}
              | _ -> {code = NodeConst1 ; haut = 1}
          in
          if Random.int 1000 < p_non then (* On met NON devant la var, action ou const *)
            {code = NodeOpNON (new_var_act_cst ()) ; haut = 1}
          else
            new_var_act_cst ()
        else (* Le noeud est recopier sans modification *)
          copie_prog pgm
      )
    | NodeOpSI (fils1,fils2,fils3) -> (* Le SI n'est jamais altr *)
        {code = NodeOpSI (cp_mut_prog fils1 proba,cp_mut_prog fils2 proba,cp_mut_prog fils3 proba) ;
         haut = 1}
    | NodeOpOU (fils1,fils2) -> 
        if Random.int 1000 < proba then (* On altre le noeud *)
          {code = NodeOpET (cp_mut_prog fils1 proba,cp_mut_prog fils2 proba) ; haut = 1}
        else (* Le noeud est recopier sans modification *)
          {code = NodeOpOU (cp_mut_prog fils1 proba,cp_mut_prog fils2 proba) ; haut = 1}
    | NodeOpET (fils1,fils2) ->
        if Random.int 1000 < proba then (* On altre le noeud *)
          {code = NodeOpOU (cp_mut_prog fils1 proba,cp_mut_prog fils2 proba) ; haut = 1}
        else (* Le noeud est recopier sans modification *)
          {code = NodeOpET (cp_mut_prog fils1 proba,cp_mut_prog fils2 proba) ; haut = 1}
    | NodeOpNON (fils1)  ->
        if Random.int 1000 < proba then (* On altre le noeud *)
          cp_mut_prog fils1 proba
        else (* Le noeud est recopier sans modification *)
          {code = NodeOpNON (cp_mut_prog fils1 proba) ; haut = pgm.haut}
;;
